/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  en Automatique.  All rights reserved.  This file is distributed    */
/*  under the terms of the GNU Library General Public License, with    */
/*  the special exception on linking described in file ../../LICENSE.  */
/*                                                                     */
/***********************************************************************/

/* $Id: text.c,v 1.15 2004/03/24 15:02:06 starynke Exp $ */

/* Patched by Jean Mouric for use with MacCaml/WinCaml: 2019/09/10 */

#include <locale.h>
#include "libgraph.h"
#include <alloc.h>
#include <string.h>
#include <fail.h>

#define MAXLENGTH 255
static int text_size = 120;
static char current_xlfd_pattern[MAXLENGTH + 1] = DEFAULT_XLFD_PATTERN;
XFontSet font_set = NULL;

static size_t text_length(char * str)
{
    setlocale(LC_CTYPE, "fr_FR.UTF-8");
    size_t lenW = mbstowcs( NULL, str, 0);
    return lenW;
}

value caml_gr_text_length(value str)
{
    size_t i = text_length(String_val(str));
    if (i == (size_t)-1) {caml_invalid_argument("text_length"); return Val_unit;}
    return Val_int((int)i);
}

static void create_font_set(char* xlfd_pattern)
{
    setlocale(LC_CTYPE, "fr_FR.UTF-8");
    char **missing_charset_list;
    int missing_charset_count;
    char *def_string;
    char str[MAXLENGTH + 1];
    sprintf(str, xlfd_pattern, text_size);
    XLockDisplay(caml_gr_display);
    font_set = XCreateFontSet(caml_gr_display, str, &missing_charset_list, &missing_charset_count, &def_string);
    XUnlockDisplay(caml_gr_display);
    if (font_set == NULL) caml_invalid_argument("font set creation failure");
}

static int check_font_name(char* name)
{
    int i = 0;
    char* p = name;
    if (*p == '-') return -1;
    char* q = name;
    while (*p) {
        if (*p == '-'){
            ++i;
            ++p;
            q = p;
            if (i > 1) return -2;
        }
        else ++p;
    }
    if (*q == 0) return -3;
    return q - name;
}

value caml_gr_set_font(value fontname)
{
    char* fntName = String_val(fontname);
    int l = strlen(fntName) + strlen("-medium-r-*-*-*-%d-*-*-*-*-*-*");
    int i = check_font_name(fntName);
    char str[MAXLENGTH + 1];
    if (i == 0){
        if (l + 3 > MAXLENGTH) caml_invalid_argument("font name too long");
        strcpy(str, "-*-");
        strcat(str, fntName);
    }
    else if (i > 0){
        if (l + 1 > MAXLENGTH) caml_invalid_argument("font name too long");
        strcpy(str, "-");
        strcat(str, fntName);
    }
    else caml_invalid_argument("font name here must have at most one '-' and no leading or trailing '-'");
    strcat(str, "-medium-r-*-*-*-%d-*-*-*-*-*-*");
    create_font_set(str);
    strcpy(current_xlfd_pattern, str);
    return Val_unit;
}

void free_font()
{
    XLockDisplay(caml_gr_display);
    if (font_set) {
       XFreeFontSet(caml_gr_display, font_set);
      font_set = NULL;
    }
    XUnlockDisplay(caml_gr_display);
}

value caml_gr_set_text_size (value sz)
{
    int size = 10 * Int_val(sz);
    if (text_size != size)
    {
        text_size = size;
        free_font();
        create_font_set(current_xlfd_pattern);
    }
    return Val_unit;
}

static int num(unsigned char c)
{
    if (c >> 7 == 0) return 0;
    if (c >> 6 == 2) return -1;
    if (c >> 5 == 6) return 1;
    if (c >> 4 == 14) return 2;
    if (c >> 3 == 30) return 3;
    return 4;
}

// test de conformité à UTF-8 d'une chaîne s
int is_utf8(const unsigned char* s)
{
    int m = 0;
    for (unsigned char* t = (unsigned char*)s; *t != '\0'; t++)
    {
        int n = num(*t);
        if (n == 4) return 0;
        if (m > 0)
        {
            if (n == -1) m--; else return 0;
        }
        else
        {
            m = n;
        }
    }
    return (m == 0);
}

// conversion d'une chaîne txt d'octets censés représenter des caractères latin1 en chaîne UTF-8
// chaque caractères UTF-8 occupent ici 1 ou 2 octets selon qu'il encode un octet < 128 ou non
// la chaîne d'octets convertie est au plus deux fois plus longue que la chaîne inintiale
/*
void to_utf8(char* txt, char* st)
{
    unsigned char *in = txt, *out=st;
    while (*in)
        if (*in<128) *out++=*in++;
        else *out++=0xc2+(*in>0xbf), *out++=(*in++&0x3f)+0x80;
    *out=0;
}
*/

void to_utf8(char* txt, char* st)
{
    unsigned char* out = st;
    for(unsigned char* in = txt; *in !='\0'; in++) {
        if (*in < 128) *out++ = *in;
        else {
            *out++ = (0xC0 | (*in >> 6));
            *out++ = (0x80 | (*in & 0x3F));
        }
    }
    *out = 0;;
}


static void caml_gr_draw_text_utf8(char * txt, int len)
//Si la chaîne txt n'est pas conforme à UTF-8 on la convertit
{
    char* st = NULL;
    int b = !is_utf8(txt);
    if (b)
    {
        st = (char*)malloc(2*strlen(txt)+1);
        if (st) to_utf8(txt, st);
        len = strlen(st);
    }
    else st = txt;
    
    if (font_set == NULL) create_font_set(current_xlfd_pattern);
    if (font_set) {
        XLockDisplay(caml_gr_display);
        setlocale(LC_CTYPE, "fr_FR.UTF-8");
        if (caml_gr_remember_modeflag) {
            Xutf8DrawString(caml_gr_display, caml_gr_bstore.win, font_set, caml_gr_bstore.gc,
                            caml_gr_x, Bcvt(caml_gr_y), st, len);
        }
        if (caml_gr_display_modeflag) {
            Xutf8DrawString(caml_gr_display, caml_gr_window.win, font_set, caml_gr_window.gc,
                            caml_gr_x, Wcvt(caml_gr_y), st, len);
            XFlush(caml_gr_display);
        }
        caml_gr_x += Xutf8TextEscapement(font_set, st, len);
        XUnlockDisplay(caml_gr_display);
    }
    if (b && st) free(st);
}


value caml_gr_draw_char(value chr)
{
    unsigned char i =  Int_val(chr);
    wchar_t str[2];
    str[0] = i;
    str[1] = 0;
    char str1[3] ;
    setlocale(LC_CTYPE, "fr_FR.UTF-8");
    int n = wcstombs(str1, str, 2);
    if (n == (size_t)-1) {caml_invalid_argument("draw_char"); return Val_unit;}
    str1[n]='\0';
    caml_gr_draw_text_utf8(str1, n);
    return Val_unit;
}

value caml_gr_draw_string(value str)
{
  caml_gr_check_open();
  caml_gr_draw_text_utf8(String_val(str), caml_string_length(str));
  return Val_unit;
}

value caml_gr_text_size(value str)
{
    int width;
    value res;
    caml_gr_check_open();
    char *s = String_val(str);
    char* st = s;
    int b = !is_utf8(s);
    if (b) {st = (char*)malloc(2*strlen(s)+1);if (st) to_utf8(s,st);}
    XLockDisplay(caml_gr_display);
    if (font_set == NULL) create_font_set(current_xlfd_pattern);
//    width = Xutf8TextEscapement(font_set, String_val(str), caml_string_length(str));
    width = Xutf8TextEscapement(font_set, st, strlen(st));
    XUnlockDisplay(caml_gr_display);
    if (b && st != s) free(st);
    res = caml_alloc_small(2, 0);
    Field(res, 0) = Val_int(width);
    Field(res, 1) = Val_int(text_size / 10);
    return res;
}
